@@ -10,6 +10,10 @@
 
 #include "postgres.h"
 
+#if PG_VERSION_NUM < 90400
+#error "PoWA requires PostgreSQL 9.4 or later"
+#endif
+
 /* For a bgworker */
 #include "miscadmin.h"
 #include "postmaster/bgworker.h"
@@ -45,8 +49,9 @@
 
 PG_MODULE_MAGIC;
 
-#define POWA_STAT_FUNC_COLS 4	/* # of cols for functions stat SRF */
-#define POWA_STAT_TAB_COLS 21	/* # of cols for relations stat SRF */
+#define POWA_STAT_FUNC_COLS	4	/* # of cols for functions stat SRF */
+#define POWA_STAT_TAB_COLS	21	/* # of cols for relations stat SRF */
+#define MIN_POWA_FREQUENCY	5000 /* minimum ms between two snapshots */
 
 typedef enum
 {
@@ -54,8 +59,10 @@ typedef enum
 	POWA_STAT_TABLE
 }	PowaStatKind;
 
-void		_PG_init(void);
-void		die_on_too_small_frequency(void);
+void			_PG_init(void);
+static bool		powa_check_frequency_hook(int *newval, void **extra, GucSource source);
+static void		compute_powa_frequency(void);
+static int64	compute_next_wakeup(void);
 
 Datum		powa_stat_user_functions(PG_FUNCTION_ARGS);
 Datum		powa_stat_all_rel(PG_FUNCTION_ARGS);
@@ -64,26 +71,92 @@ static Datum powa_stat_common(PG_FUNCTION_ARGS, PowaStatKind kind);
 PG_FUNCTION_INFO_V1(powa_stat_user_functions);
 PG_FUNCTION_INFO_V1(powa_stat_all_rel);
 
-static void powa_main(Datum main_arg);
+#if (PG_VERSION_NUM >= 90500)
+void powa_main(Datum main_arg) pg_attribute_noreturn();
+#else
+void powa_main(Datum main_arg) __attribute__((noreturn));
+#endif
+
 static void powa_sighup(SIGNAL_ARGS);
+static void powa_process_sighup(void);
 
-static int	powa_frequency;
-static int	min_powa_frequency = 5000;
-static int	powa_retention;
-static int	powa_coalesce;
-static char *powa_database = NULL;
-static char *powa_ignored_users = NULL;
-static bool powa_debug = false;
+static instr_time	last_start;					/* last snapshot start */
 
-void
-die_on_too_small_frequency(void)
+static int			powa_frequency;				/* powa.frequency GUC */
+static instr_time	time_powa_frequency;		/* same in instr_time format */
+static bool			force_snapshot = false;		/* used to force snapshot after enabling powa */
+static int			powa_retention;				/* powa.retention GUC */
+static int			powa_coalesce;			 	/* powa.coalesce GUC */
+static char		   *powa_database = NULL;	 	/* powa.database GUC */
+static char 	   *powa_ignored_users = NULL;	/* powa.ignored_users GUC */
+static bool			powa_debug = false;			/* powa.debug GUC */
+
+/* flags set by signal handlers */
+static volatile sig_atomic_t got_sighup = false;
+
+static bool
+powa_check_frequency_hook(int *newval, void **extra, GucSource source)
 {
-	if (powa_frequency > 0 && powa_frequency < min_powa_frequency)
+	if (*newval >= MIN_POWA_FREQUENCY || *newval == -1)
+		return true;
+
+	return false;
+}
+
+static void
+compute_powa_frequency(void)
+{
+	int local_frequency = powa_frequency;
+
+	/*
+	 * If PoWA is deactivated, artitrarily set a sleep time of one hour to save
+	 * resources.  The actual sleep time is not problematic since the
+	 * reactivation can only be done with a sighup, which will set the latch
+	 * used for the sleep.
+	 */
+	if (powa_frequency == -1)
+		local_frequency = 3600000;
+
+	/* Initialize time_powa_frequency to do maths with it */
+#ifndef WIN32
+#ifdef HAVE_CLOCK_GETTIME
+	time_powa_frequency.tv_sec = local_frequency / 1000; /* Seconds */
+	time_powa_frequency.tv_nsec = 0;
+#else		/* !HAVE_CLOCK_GETTIME */
+	time_powa_frequency.tv_sec = local_frequency / 1000; /* Seconds */
+	time_powa_frequency.tv_usec = 0;
+#endif		/* HAVE_CLOCK_GETTIME */
+
+#else		/* WIN32 */
+	time_powa_frequency.QuadPart = local_frequency / 1000 * GetTimerFrequency();
+#endif		/* WIN32 */
+}
+
+static int64
+compute_next_wakeup(void)
+{
+	instr_time time_to_wait, now;
+
+	/*
+	 * If powa was deactivated and is now reactivated, we force an immediate
+	 * snapshot, and start usual interval sleep.  In order to do that, we have
+	 * to setup a new last_start reference to now minus powa_frequency so that
+	 * the next calls to this function will return expected sleep time.
+	 */
+	if (force_snapshot)
 	{
-		elog(LOG, "POWA frequency cannot be smaller than %i milliseconds",
-			 min_powa_frequency);
-		exit(1);
+		force_snapshot = false;
+		INSTR_TIME_SET_CURRENT(last_start);
+		INSTR_TIME_SUBTRACT(last_start, time_powa_frequency);
+		return 0;
 	}
+
+	memcpy(&time_to_wait, &last_start, sizeof(last_start));
+	INSTR_TIME_ADD(time_to_wait, time_powa_frequency);
+	INSTR_TIME_SET_CURRENT(now);
+	INSTR_TIME_SUBTRACT(time_to_wait, now);
+
+	return INSTR_TIME_GET_MICROSEC(time_to_wait);
 }
 
 void
@@ -105,7 +178,10 @@ _PG_init(void)
 							300000,
 							-1,
 							INT_MAX / 1000,
-							PGC_SUSET, GUC_UNIT_MS, NULL, NULL, NULL);
+							PGC_SUSET, GUC_UNIT_MS,
+							powa_check_frequency_hook,
+							NULL,
+							NULL);
 
 	DefineCustomIntVariable("powa.coalesce",
 							"Defines the amount of records to group together in the table (more compact)",
@@ -140,6 +216,8 @@ _PG_init(void)
 							   &powa_debug,
 							   false, PGC_USERSET, 0, NULL, NULL, NULL);
 
+	EmitWarningsOnPlaceholders("powa");
+
 	/*
 	 * Register the worker processes
 	 */
@@ -147,7 +225,12 @@ _PG_init(void)
 		BGWORKER_SHMEM_ACCESS | BGWORKER_BACKEND_DATABASE_CONNECTION;
 	worker.bgw_start_time = BgWorkerStart_RecoveryFinished;		/* Must write to the
 																 * database */
+#if (PG_VERSION_NUM >= 100000)
+	snprintf(worker.bgw_library_name, BGW_MAXLEN, "powa");
+	snprintf(worker.bgw_function_name, BGW_MAXLEN, "powa_main");
+#else
 	worker.bgw_main = powa_main;
+#endif
 	snprintf(worker.bgw_name, BGW_MAXLEN, "powa");
 	worker.bgw_restart_time = 10;
 	worker.bgw_main_arg = (Datum) 0;
@@ -158,16 +241,16 @@ _PG_init(void)
 }
 
 
-static void
+void
 powa_main(Datum main_arg)
 {
 	char	   *query_snapshot = "SELECT powa_take_snapshot()";
 	static char *query_appname = "SET application_name = 'PoWA collector'";
-	instr_time	begin;
-	instr_time	end;
-	long		time_to_wait;
+	int64		us_to_wait; /* Should be uint64 per postgresql's spec, but we
+							   may have negative result, in our tests */
 
-	die_on_too_small_frequency();
+	/* check powa_frequency validity, and if powa is enabled */
+	compute_powa_frequency();
 
 	/*
 	 * Set up signal handler, then unblock signals
@@ -176,23 +259,15 @@ powa_main(Datum main_arg)
 
 	BackgroundWorkerUnblockSignals();
 
-	/*
-	 * We only connect when powa_frequency >0. If not, powa has been
-	 * deactivated
-	 */
-	if (powa_frequency < 0)
-	{
-		elog(LOG, "POWA is deactivated (powa.frequency = %i), exiting",
-			 powa_frequency);
-		exit(1);
-	}
-	/* We got here: it means powa_frequency > 0. Let's connect */
+	/* Define the snapshot reference time when the bgworker start */
+	INSTR_TIME_SET_CURRENT(last_start);
 
-
-	/*
-	 * Connect to POWA database
-	 */
+	/* Connect to POWA database */
+#if PG_VERSION_NUM >= 110000
+	BackgroundWorkerInitializeConnection(powa_database, NULL, 0);
+#else
 	BackgroundWorkerInitializeConnection(powa_database, NULL);
+#endif
 
 	elog(LOG, "POWA connected to database %s", quote_identifier(powa_database));
 
@@ -212,81 +287,122 @@ powa_main(Datum main_arg)
 	/*------------------
 	 * Main loop of POWA
 	 * We exit from here if:
-	 *	 - we got a SIGINT/SIGTERM (default bgworker sig handlers)
+	 *	 - we got a SIGINT (default bgworker sig handlers)
 	 *	 - powa.frequency becomes < 0 (change config and SIGHUP)
 	 */
-	while (true)
+	for (;;)
 	{
-		/*
-		 * We can get here with a new value of powa_frequency because of a
-		 * reload. Let's suicide to disconnect if this value is <0
-		 */
-		if (powa_frequency < 0)
+		/* Check if a SIGHUP has been received */
+		powa_process_sighup();
+
+		if (powa_frequency != -1)
 		{
-			elog(LOG, "POWA exits to disconnect from the database now");
-			exit(1);
+			set_ps_display("snapshot", false);
+			SetCurrentStatementStartTimestamp();
+			StartTransactionCommand();
+			SPI_connect();
+			PushActiveSnapshot(GetTransactionSnapshot());
+			pgstat_report_activity(STATE_RUNNING, query_snapshot);
+			SPI_execute(query_snapshot, false, 0);
+			pgstat_report_activity(STATE_RUNNING, query_appname);
+			SPI_execute(query_appname, false, 0);
+			SPI_finish();
+			PopActiveSnapshot();
+			CommitTransactionCommand();
+			pgstat_report_stat(false);
+			pgstat_report_activity(STATE_IDLE, NULL);
+			set_ps_display("idle", false);
 		}
 
-		/*
-		 * let's store the current time. It will be used to calculate a quite
-		 * stable interval between each measure
-		 */
-		set_ps_display("snapshot", false);
-		INSTR_TIME_SET_CURRENT(begin);
-		ResetLatch(&MyProc->procLatch);
-		SetCurrentStatementStartTimestamp();
-		StartTransactionCommand();
-		SPI_connect();
-		PushActiveSnapshot(GetTransactionSnapshot());
-		pgstat_report_activity(STATE_RUNNING, query_snapshot);
-		SPI_execute(query_snapshot, false, 0);
-		pgstat_report_activity(STATE_RUNNING, query_appname);
-		SPI_execute(query_appname, false, 0);
-		SPI_finish();
-		PopActiveSnapshot();
-		CommitTransactionCommand();
-		pgstat_report_stat(false);
-		pgstat_report_activity(STATE_IDLE, NULL);
-		set_ps_display("idle", false);
-		INSTR_TIME_SET_CURRENT(end);
-		INSTR_TIME_SUBTRACT(end, begin);
-
-		/*
-		 * Wait powa.frequency, compensate for work time of last snapshot
-		 */
-
-		/*
-		 * If we got off schedule (because of a compact or delete, just do
-		 * another operation right now
-		 */
-		time_to_wait = powa_frequency - INSTR_TIME_GET_MILLISEC(end);
-		elog(DEBUG1, "Waiting for %li milliseconds", time_to_wait);
-		if (time_to_wait > 0)
+		/* sleep loop */
+		for (;;)
 		{
 			StringInfoData buf;
 
+			/* Check if a SIGHUP has been received */
+			powa_process_sighup();
+
+			/*
+			 * Compute if there is still some time to wait (we could have been
+			 * woken up by a latch, or snapshot took more than frequency)
+			 */
+			us_to_wait = compute_next_wakeup();
+			if (us_to_wait <= 0)
+				break;
+
+			/* Tell the world we are waiting */
+			elog(DEBUG1, "Waiting for %li milliseconds", us_to_wait/1000);
 			initStringInfo(&buf);
-
 			appendStringInfo(&buf, "-- sleeping for %li seconds",
-							 time_to_wait / 1000);
-
+							 us_to_wait / 1000000);
 			pgstat_report_activity(STATE_IDLE, buf.data);
-
 			pfree(buf.data);
 
+			/* sleep */
 			WaitLatch(&MyProc->procLatch,
 					  WL_LATCH_SET | WL_TIMEOUT | WL_POSTMASTER_DEATH,
-					  time_to_wait);
-		}
-	}
+					  us_to_wait/1000
+#if PG_VERSION_NUM >= 100000
+					  ,PG_WAIT_EXTENSION
+#endif
+					  );
+			ResetLatch(&MyProc->procLatch);
+		} /* end of sleep loop */
+
+		/*
+		 * We've stop waiting. Let's increment the snapshot reference time to
+		 * it's ideal target, not to now, so errors don't add up
+		 */
+		INSTR_TIME_ADD(last_start, time_powa_frequency);
+	} /* end of snapshot loop */
 }
 
 
+/*
+ * Signal handler for SIGHUP
+ *		Set a flag to tell the main loop to reread the config file, do sanity
+ *		check, recompute the frequency and set our latch to wake it up.
+ */
 static void
 powa_sighup(SIGNAL_ARGS)
 {
-	ProcessConfigFile(PGC_SIGHUP);
-	die_on_too_small_frequency();
+	int			save_errno = errno;
+
+	got_sighup = true;
+
+	if (MyProc)
+		SetLatch(&MyProc->procLatch);
+
+	errno = save_errno;
+}
+
+/*
+ * Do all the needed work if a SIGHUP has been received
+ *		- reread the config file
+ *		- exit the bgworker if the frequency is invalid
+ *		- compute the time_powa_frequency var
+ */
+static void
+powa_process_sighup(void)
+{
+	if (got_sighup)
+	{
+		int old_powa_frequency = powa_frequency;
+
+		got_sighup = false;
+
+		ProcessConfigFile(PGC_SIGHUP);
+		/* setup force_snapshot if powa is reactivated */
+		if (old_powa_frequency == -1 && powa_frequency != -1)
+		{
+			elog(LOG, "PoWA is activated");
+			force_snapshot = (old_powa_frequency == -1 && powa_frequency != -1);
+		}
+		else if (old_powa_frequency != -1 && powa_frequency == -1)
+			elog(LOG, "PoWA is deactivated");
+
+		compute_powa_frequency();
+	}
 }
 
 Datum
@@ -371,6 +487,10 @@ powa_stat_common(PG_FUNCTION_ARGS, PowaStatKind kind)
 	 * - and finally clear again the statistics cache, to make sure any further
 	 *	 statement in the transaction will see the data related to the right
 	 *	 database.
+	 *
+	 *	 The pgstat_fetch_stat_dbentry() has to be done inside a PG_TRY bloc so
+	 *	 we make sure that MyDatabaseId is restored and the statistics cache is
+	 *	 cleared if an error happens during the call.
 	 */
 
 	pgstat_clear_snapshot();
@@ -378,11 +498,23 @@ powa_stat_common(PG_FUNCTION_ARGS, PowaStatKind kind)
 	backend_dbid = MyDatabaseId;
 	MyDatabaseId = dbid;
 
-	dbentry = pgstat_fetch_stat_dbentry(dbid);
+	PG_TRY();
+	{
+		dbentry = pgstat_fetch_stat_dbentry(dbid);
+	}
+	PG_CATCH();
+	{
+		MyDatabaseId = backend_dbid;
+		pgstat_clear_snapshot();
+
+		PG_RE_THROW();
+	}
+	PG_END_TRY();
 
 	MyDatabaseId = backend_dbid;
 
-	if (dbentry != NULL && dbentry->functions != NULL)
+	if (dbentry != NULL && dbentry->functions != NULL &&
+		dbentry->tables != NULL)
 	{
 		switch (kind)
 		{
